""" SimpleScriptEngine for ScriptEngine.

SimpleScriptEngine is a simplistic script engine. It runs all jobs sequentially
on the local machine, without consideration of job contexts.
"""

import sys
import logging

from scriptengine.tasks.core import Task
from scriptengine.jobs import Job
from scriptengine.exceptions import ScriptEngineError, \
                                    ScriptEngineStopException


class SimpleScriptEngine:
    '''Simplistic script engine for ScriptEngine.
    '''
    def _guarded_run(self, item, context):
        try:
            item.run(context)
        except ScriptEngineStopException:
            self.log_info(
                'STOPPING ScriptEngine instance upon request'
            )
            sys.exit()

    def run(self, script, context):

        for script_item in script if isinstance(script, list) else [script]:

            if isinstance(script_item, Task):
                self._guarded_run(script_item, context)
            elif isinstance(script_item, Job):
                if script_item.when(context):

                    # Loop setup with loop var collision detection
                    none_loop = object()  # just a unique object
                    loop_var, loop_iter = script_item.loop(context)
                    old_loop_var = context.get(loop_var, None)
                    if old_loop_var:
                        self.log_warning(
                            'Loop variable collision for \'{loop_var}\''
                        )
                    # Run loop (at least once)
                    for loop_item in loop_iter or [none_loop]:

                        if loop_item is not none_loop:
                            context[loop_var] = loop_item

                        # Run task/job for one loop iteration
                        for todo in script_item.todo:
                            if isinstance(todo, Task):
                                self._guarded_run(todo, context)
                            elif isinstance(todo, Job):
                                # recurse for jobs
                                self.run(todo, context)
                            else:
                                self.log_error(
                                    'Invalid item (neither Task nor Job) '
                                    f'of type \'{type(script_item)}\''
                                )
                                raise ScriptEngineError
                    # Remove or restore loop var
                    if old_loop_var:
                        context[loop_var] = old_loop_var
                    else:
                        context.pop(loop_var, None)
                else:
                    self.log_debug(
                        f'Not executing <{script_item.shortid}> because '
                        'when clause evaluates false')
            else:
                self.log_error(
                    'Invalid item (neither Task nor Job) '
                    f'of type \'{type(script_item)}\''
                )
                raise ScriptEngineError

    def _log(self, level, msg):
        logger = 'se.instance.'+self.__class__.__name__.lower()
        logging.getLogger(logger).log(level, msg)

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



"""ScriptEngine YAML parsing
"""

import yaml
import dateutil.rrule
import logging

import scriptengine.tasks.core.loader
from scriptengine.jobs import Job
from scriptengine.exceptions import ScriptEngineParseFileError, \
                                    ScriptEngineParseScriptError, \
                                    ScriptEngineParseYAMLError


class NoParseYamlString(str):
    """This represents a string that should not be YAML-parsed"""


class NoParseJinjaString(str):
    """This represents a string that should not be parsed with Jinja2"""


class NoParseString(NoParseYamlString, NoParseJinjaString):
    """This represents a string that should not be parsed at all, neither YAML
       or Jinja2"""


def string_class_constructor(derived_string_class):
    """YAML constructor factory. The constructors convert their values to a
       string and return an object of a specific class (derived from str). The
       class membership is later used to limit the Jinja2/YAML parsing during
       the processing of ScriptEngine scripts.
    """
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return derived_string_class(str(value))
    return constructor


yaml.add_constructor(u'!noparse',
                     string_class_constructor(NoParseString))
yaml.add_constructor(u'!noparse_yaml',
                     string_class_constructor(NoParseYamlString))
yaml.add_constructor(u'!noparse_jinja',
                     string_class_constructor(NoParseJinjaString))


def rrule_constructor(loader, node):
    """A YAML constructor that can parse RFC5545 rrules
    """
    value = loader.construct_scalar(node)
    rrule = dateutil.rrule.rrulestr(value)
    return rrule


yaml.add_constructor(u'!rrule', rrule_constructor)


def parse(data):
    """
    Args:
        data (dict or list of dicts): Data structure that holds the script
            description, i.e. job and task specifications. See the simple
            ScriptEngine grammar in the documentation.

    Returns:
        A scriptengine.task.Task, a scriptengine.jobs.Job, or a list of
        tasks/jobs.
    """

    if not data:
        return []

    tasks = scriptengine.tasks.core.loader.load_and_register()
    jobs = {"do"}

    # Recursively parse lists of tasks/jobs
    if isinstance(data, list):
        return [parse(item) for item in data]

    # Parse single task/job
    try:
        keys = (jobs | tasks.keys()) & data.keys()
    except AttributeError:
        logging.getLogger('se.yaml').error(
            'Expected YAML dictionary, got '
            f'"{type(data).__name__}: {data}"'
        )
        raise ScriptEngineParseScriptError
    if not keys:
        logging.getLogger('se.yaml').error(
            'Found unknown task name(s): '
            f'\"{", ".join(data.keys())}\"'
        )
        raise ScriptEngineParseScriptError
    if len(keys) > 1:
        logging.getLogger('se.yaml').error(
            'Found ambiguous task names: '
            f'\"{", ".join(data.keys())}\"'
        )
        raise ScriptEngineParseScriptError

    # There is exactly one key, get it
    key = keys.pop()

    if key in tasks:

        if '.' not in key:
            logging.getLogger('se.yaml').warn(
                f'Deprecation warning while processing task "{key}": '
                'The use of task names without dots (i.e. without a namespace)'
                f' is deprecated! This task defaults to "base.{key}", but the '
                'default will be removed in the future and an error will '
                'occur instead.'
            )

        if len(data) == 1:  # Simple task (no when clause or loop)
            return tasks[key](data[key])
        # Job made from single task with when/loop
        job = Job(when=data.get("when"), loop=data.get("loop"))
        job.append(parse({key: data[key]}))
        return job

    job = Job(when=data.get("when"), loop=data.get("loop"))
    for item in data[key]:
        job.append(parse(item))
    return job


def parse_file(filename):
    """Reads a ScriptEngine script from a YAML file.

    Args:
        filename (str): Path to YAML file

    Returns:
        A scriptengine.task.Task, a scriptengine.jobs.Job, or a list of
        tasks/jobs.
    """
    try:
        with open(filename) as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
    except (FileNotFoundError,
            PermissionError,
            IsADirectoryError) as e:
        logging.getLogger('se.yaml').error(
            f'Could not read script file: {e}'
        )
        raise ScriptEngineParseFileError
    except yaml.YAMLError as e:
        logging.getLogger('se.yaml').error(
            f'Could not parse script file: {e}'
        )
        raise ScriptEngineParseYAMLError
    return parse(data)



"""Include task for ScriptEngine

   The Include task reads a YAML file, parses the content to create a
   ScriptEngine script and lets the active ScriptEngine instance execute it.
"""

import os

import scriptengine.yaml

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.exceptions import ScriptEngineTaskRunError


class Include(Task):

    _required_arguments = ('src', )

    def __init__(self, arguments):
        Include.check_arguments(arguments)
        super().__init__(arguments)

    def __str__(self):
        return f'Include: {self.src}'

    @timed_runner
    def run(self, context):
        inc_file = self.getarg('src', context)
        self.log_info(f'Include script from {inc_file}')

        search_path = ('.', context.se.cli.cwd) + context.se.cli.script_path
        self.log_debug(f'Searching in path: {search_path}')

        for directory in search_path:
            inc_file_path = os.path.join(directory, inc_file)
            self.log_debug(f'Searching for include file at "{inc_file_path}"')
            if os.path.isfile(inc_file_path):
                self.log_debug(f'Found include file at "{inc_file_path}"')
                break
        else:
            if self.getarg('ignore_not_found', default=False):
                self.log_warning(f'Include file "{inc_file}" not found.')
                return
            msg = f'Include file "{inc_file}" not found'
            self.log_error(msg)
            raise ScriptEngineTaskRunError(msg)

        script = scriptengine.yaml.parse_file(inc_file_path)

        inc_file_dir = os.path.dirname(os.path.abspath(inc_file_path))
        if inc_file_dir not in context.se.cli.script_path:
            context.se.cli.script_path += inc_file_dir,
        self.log_debug(f'New script_path: {context.se.cli.script_path}')

        self.log_debug(f'Execute script from "{inc_file}"')
        context.se.instance.run(script, context)
        self.log_debug(f'Finished executing script from "{inc_file}"')



"""ScriptEngine timing for tasks: provides timing of Task classes .run() method
"""

import time
import functools


def timed_runner(func):
    """Wrapper for run() functions of SE tasks. To be used as decorator as follows:

        from scriptengine.task_timing import timed_runner
        class Foo(Task):
            # ...
            @timed_runner
            def run(self, context):
                # ...
    """

    @functools.wraps(func)
    def wrap_timed(self, context):

        try:
            mode = context.se.tasks.timing.mode
        except AttributeError:
            mode = False

        if mode in ('basic', 'classes', 'instances'):

            # timed function call
            start_tic = time.perf_counter()
            func_return_value = func(self, context)
            elapsed_time = time.perf_counter() - start_tic

            # logging
            if context.se.tasks.timing.logging == 'info':
                self.log_info(f'Elapsed time: {elapsed_time:0.4f} seconds')
            elif context.se.tasks.timing.logging == 'debug':
                self.log_debug(f'Elapsed time: {elapsed_time:0.4f} seconds')

            # update timers
            if mode in ('classes', 'instances'):
                timers = context['se']['tasks']['timing']['timers']

                class_t = timers.setdefault('classes', {})
                class_t.setdefault(self.__class__.__name__, 0)
                class_t[self.__class__.__name__] += elapsed_time

                if mode == 'instances':
                    instance_t = timers.setdefault('instances', {})
                    instance_t.setdefault(self.id, 0)
                    instance_t[self.id] += elapsed_time

            return func_return_value

        # if no timing, just return the function
        return func(self, context)

    return wrap_timed



""" ScriptEngine jobs.

Jobs are the second level (above tasks) work unit in ScriptEngine. Jobs are
made of tasks and may include loops, conditionals, and a context. Job lists can
be given to a ScriptEngine, for execution.
"""

import logging
import uuid
import ast

from scriptengine.jinja import render as j2render
from scriptengine.exceptions import ScriptEngineParseError


class Job:

    def __init__(self, when=None, loop=None):
        self._identifier = uuid.uuid4()
        self._todo = []
        self._when = when

        if loop is None:
            self._loop_var = None
            self._loop_iter = None
        elif isinstance(loop, dict):
            self._loop_var = loop.get("with", "item")
            self._loop_iter = loop.get("in", None)
        else:  # loop is assumed list
            self._loop_var = "item"
            self._loop_iter = loop

        self.log_debug(
                f'Create Job'
                f'{" when "+self._when if self._when else ""}'
                f'{" with "+str(self._loop_var) if self._loop_var else ""}'
                f'{" in "+str(self._loop_iter) if self._loop_iter else ""}')

    @property
    def id(self):
        return self._identifier

    @property
    def shortid(self):
        return self._identifier.hex[:10]

    @property
    def todo(self):
        return iter(self._todo)

    def when(self, context):
        return (self._when is None
                or j2render(self._when, context, boolean=True))

    def loop(self, context):
        loop_iter = j2render(self._loop_iter, context)
        if isinstance(loop_iter, str):
            try:
                loop_iter = ast.literal_eval(loop_iter or 'None')
            except SyntaxError:
                self.log_error(
                    'Syntax error while evaluating loop '
                    f'expression \'{loop_iter}\''
                )
                raise ScriptEngineParseError
        return self._loop_var, loop_iter

    def append(self, todo):
        todo_as_list = todo if isinstance(todo, list) else [todo]
        self.log_debug(
                f"Append: {','.join([str(t.id) for t in todo_as_list])} to")
        self._todo.extend(todo_as_list)

    def _log(self, level, msg):
        logging.getLogger('se.job').log(level, msg, extra={'id': self.shortid})

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



'''ScriptEngine command line interface (cli).

The se command provides a command line interface to ScriptEngine, allowing for
the creation of scripts from YAML files and runnig the scripts via the
SimpleScriptEngine.
'''

__copyright__ = '''
Copyright 2019, 2020 Uwe Fladrich

This file is part of ScriptEngine.

ScriptEngine is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ScriptEngine is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ScriptEngine. If not, see <https://www.gnu.org/licenses/>.
'''
__license__ = 'GPLv3+'


import os
import argparse
import logging

from attrdict import AttrDict

import scriptengine.version
import scriptengine.helpers.terminal_colors
import scriptengine.logging

from scriptengine.scripts import SimpleScriptEngine
from scriptengine.exceptions import ScriptEngineParseError, \
                                    ScriptEngineParseFileError

__version__ = scriptengine.version.__version__


def parse_cmd_line_args():
    '''Parses the command line arguments with argparse
    '''
    arg_parser = argparse.ArgumentParser(
                    description='ScriptEngine command line tool',
                    epilog='Available ScriptEngine tasks: ' +
                           ', '.join(scriptengine.tasks.core.loader.load().keys()))
    arg_parser.add_argument('-V', '--version',
                            help='show ScriptEngine version and exit',
                            action='version',
                            version=__version__)
    arg_parser.add_argument('--loglevel',
                            help='The minimum level of log messages that is '
                                 'displayed. For verbose output, use '
                                 '"debug". For minimal output, use "error" '
                                 'or "critical".',
                            choices=['debug',
                                     'info',
                                     'warning',
                                     'error',
                                     'critical'])
    arg_parser.add_argument('--nocolor',
                            help='do not use colored terminal output',
                            action='store_true')
    arg_parser.add_argument('files',
                            help='YAML file(s) to read',
                            nargs='+')

    return arg_parser.parse_args()


def parse_files(logger, files):
    '''Parses files and returns a script (list of tasks/jobs).
    '''
    script = []
    for fname in files:
        # next_script is a Task or Job or list of either
        next_script = scriptengine.yaml.parse_file(fname)
        if isinstance(next_script, list):
            script.extend(next_script)
        else:
            script.append(next_script)

    return script


def main():
    '''ScriptEngine command line tool

       The context is passed to the ScriptEngine instance and, later, to all
       tasks. Some context information is already set here in the se command
       line tool:
         context.se.cli.cwd         - the original working directory
         context.se.cli.script_path - list of paths for all scripts given at
                                      the command line (used as search path
                                      for includes, for example)
         context.se.instance        - a reference to the ScriptEngine instance
                                      that executes the task
         context.se.tasks.timing    - task timing information
                                      (defaults to no timing)
    '''

    # Parse command line arguments
    parsed_args = parse_cmd_line_args()

    # Choose colored output unless --nocolor
    scriptengine.helpers.terminal_colors.set_theme(
        'none' if parsed_args.nocolor else 'standard'
    )

    # Configure logging with level given by --loglevel
    # or by default with logging.INFO
    scriptengine.logging.configure(
        logging.getLevelName(parsed_args.loglevel.upper())
        if parsed_args.loglevel else logging.INFO
    )
    logger = logging.getLogger('se.cli')
    logger.info('Logging configured and started')

    # Log all loaded tasks if debugging
    if logger.level <= logging.DEBUG:
        for name, task in scriptengine.tasks.core.loader.load().items():
            logger.debug(
                f'Loaded task {name}: '
                f'{task.__name__} '
                f'from {task.__module__}'
            )

    # Create script by parsing the files given on the command line
    try:
        script = parse_files(logger, parsed_args.files)
    except ScriptEngineParseFileError:
        logger.critical('Could not read all script files')
        return os.EX_NOINPUT
    except ScriptEngineParseError:
        logger.critical('Could not parse all script files')
        return os.EX_DATAERR

    # Note that the following removal of duplicates does not guarantee the
    # order of entries before Python 3.7! We'll accept this in favour of the
    # easier implementation.
    script_path = tuple(dict.fromkeys(
                        (os.path.dirname(file) for file in parsed_args.files)))

    # Initialise the context as AttrDict,
    # i.e. allowing *read* access as, for example, 'context.se.cli.cwd'
    context = AttrDict({
        'se': {
            'cli': {
                'cwd': os.getcwd(),
                'script_path': script_path,
            },
            'tasks': {
                'timing': {
                    'mode': None,
                    'logging': None,
                    'timers': {},
                },
            },
            'instance': SimpleScriptEngine(),
        }
    })

    # Call ScriptEngine instance to run the script
    context.se.instance.run(script, context)

    return os.EX_OK



"""Getenv task for ScriptEngine."""

import os

from scriptengine.tasks.core import Task, timed_runner


class Getenv(Task):
    """ Getenv task, reads environment variable

    The Getenv run() method reads an environmane variable and sets a
    ScriptEngine configuration parameter.

    Args:
        dictionary (dict): Must at least contain the following keys:
            - name: Name of the environment variable
            - set: Name of the ScriptEngine config parameter that is set to
                the value of the environment variable (if it exists)
    """
    _required_arguments = ('name', 'set', )

    def __init__(self, arguments):
        Getenv.check_arguments(arguments)
        super().__init__(arguments)

    def __str__(self):
        return f"Getenv: ENV{self.name} --> {self.set}"

    @timed_runner
    def run(self, context):
        self.log_info(f"ENV{self.name} --> {self.set}")
        env = os.environ.get(self.getarg('name', context))
        if env:
            context[self.getarg('set', context)] = env



"""Find task for ScriptEngine"""

import os
import fnmatch

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.exceptions import ScriptEngineTaskArgumentInvalidError


class Find(Task):
    """Find task, finds files or directories by name patterns
    """
    _required_arguments = ('path', )

    def __init__(self, arguments):
        Find.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        path = os.path.normpath(self.getarg('path', context))
        path_depth = path.count(os.path.sep)

        pattern = self.getarg('pattern', context, default='*')

        find_type = self.getarg('type', context, default='file')
        if find_type not in ('file', 'dir'):
            msg = (f'Invalid "type" argument '
                   f'(must be either "file" or "dir"): {find_type}')
            self.log_error(msg)
            raise ScriptEngineTaskArgumentInvalidError(msg)

        max_depth = self.getarg('depth', context, default=-1)
        try:
            max_depth = max(-1, int(max_depth))
        except (ValueError, TypeError):
            msg = f'Invalid "depth" argument (not an integer): {max_depth}'
            self.log_error(msg)
            raise ScriptEngineTaskArgumentInvalidError(msg)

        self.log_info(f'Find {find_type} with pattern "{pattern}" '
                      f'in {path} (with max depth={max_depth})')
        result = []
        for root, dirs, files in os.walk(path):
            if find_type == 'file':
                files_or_dirs = files
            else:
                files_or_dirs = dirs
            for name in files_or_dirs:
                if fnmatch.fnmatch(name, pattern):
                    result.append(os.path.normpath(os.path.join(root, name)))
            if (max_depth >= 0 and
                    root.count(os.path.sep) >= path_depth+max_depth):
                del dirs[:]
        if result:
            self.log_debug(f'Found: {result}')
        else:
            self.log_debug('Nothing found')
        result_key = self.getarg('set', context, default='result')
        self.log_debug(f'Store result under context key "{result_key}"')
        context[result_key] = result



"""TaskTimer for ScriptEngine."""

from scriptengine.tasks.core import Task
from scriptengine.exceptions import ScriptEngineTaskRunError


class TaskTimer(Task):
    """TaskTimer, controls ScriptEngine task timing feature

       Arguments:
         mode (required):   one of
                            False:       Timing is switched off.
                            'basic':     Each task is timed, log messages are
                                         written according to 'logging'
                                         argument.
                            'classes':   As for 'basic', plus times are
                                         accumulated for task classes.
                            'instances': As for 'classes', plus times are
                                         accumulated for each individual task
                                         instance.
        logging (optional): one of
                            False:   No time logging after each task. Does not
                                     affect statistic collection.
                            'info':  Logging to the info logger.
                            'debug': Logging to the debug logger.
    """
    _required_arguments = ('mode', )

    def __init__(self, arguments):
        TaskTimer.check_arguments(arguments)
        super().__init__(arguments)

    def run(self, context):

        mode = self.getarg('mode', context)
        if mode not in (False, 'basic', 'classes', 'instances'):
            msg = ('Unknown mode in TaskTimer: must be one of "off", "basic", '
                   f'"classes", "instances", not "{mode}"')
            self.log_error(msg)
            raise ScriptEngineTaskRunError(msg)

        if mode:
            logging = self.getarg('logging', context, default=False)
            if logging not in (False, 'info', 'debug'):
                msg = ('Unknown logging mode in TaskTimer: must be one of '
                       f'"off", "info", "debug", not "{logging}"')
                self.log_error(msg)
                raise ScriptEngineTaskRunError(msg)
            self.log_info(f'Task timing activated in mode "{mode}" and '
                          f'logging to "{logging}"')
        else:
            logging = False
            self.log_info('Task timing is switched off')

        context['se']['tasks']['timing']['mode'] = mode
        context['se']['tasks']['timing']['logging'] = logging



"""ScriptEngine task module.

Provides the base class for all tasks.
"""

import logging
import uuid
import yaml

import scriptengine.jinja
import scriptengine.yaml

from scriptengine.exceptions import ScriptEngineTaskArgumentError, \
                                    ScriptEngineTaskArgumentInvalidError, \
                                    ScriptEngineTaskArgumentMissingError


_SENTINEL = object()


class Task:

    _reg_name = None
    _invalid_arguments = ('run', 'id', )

    @classmethod
    def check_arguments(cls, arguments):
        """Checks arguments against class members '_invalid_arguments' and
        '_required_arguments', if present. Returns None or throws either
        ScriptEngineTaskArgumentInvalidError or
        ScriptEngineTaskArgumentMissingError.
        """
        for name in arguments:
            if name in getattr(cls, '_invalid_arguments', ()):
                logmsg = (f'Invalid argument "{name}" found while '
                          f'trying to create {cls.__name__} task')
                logging.getLogger('se.task').error(logmsg, extra={'id': None})
                raise ScriptEngineTaskArgumentInvalidError(logmsg)
        for name in getattr(cls, '_required_arguments', ()):
            if name not in arguments:
                logmsg = (f'Missing required argument "{name}" while '
                          f'trying to create {cls.__name__} task')
                logging.getLogger('se.task').error(logmsg, extra={'id': None})
                raise ScriptEngineTaskArgumentMissingError(logmsg)

    def __init__(self, arguments=None):

        self._identifier = uuid.uuid4()

        if arguments is not None:
            try:
                Task.check_arguments(arguments)
            except ScriptEngineTaskArgumentError as e:
                self.log_error(e)
                raise
            for name, value in arguments.items():
                if hasattr(self, name):
                    self.log_error(f'Invalid task argument (reserved): {name}')
                    raise ScriptEngineTaskArgumentInvalidError(
                                   f'Invalid task argument (reserved): {name}')
                setattr(self, name, value)
        self.log_debug(f'Created task: {self}')

    @classmethod
    def register_name(cls, name):
        cls._reg_name = name

    @property
    def reg_name(self):
        return self._reg_name or \
               f'{self.__class__.__module__}:{self.__class__.__name__}'

    @property
    def id(self):
        return self._identifier

    @property
    def shortid(self):
        return self._identifier.hex[:10]

    def __repr__(self):
        params = {key: val for key, val in self.__dict__.items()
                  if not (isinstance(key, str) and key.startswith('_'))}
        params_list = f'{", ".join([f"{k}={params[k]}" for k in params])}'
        return f'{self.__class__.__name__}({params_list})'

    def run(self, context):
        raise NotImplementedError('Base class function Task.run() '
                                  'must not be called')

    def getarg(self, name, context={},
               parse_jinja=True, parse_yaml=True, default=_SENTINEL):
        """Returns the value of argument 'name'.
           The argument value is parsed with Jinja2 and the given context,
           unless 'parse_jinja' is False. The result is parsed once more with
           the YAML parser in order to get a correctly typed result. If
           parse_yaml is not True, the extra YAML parsing is skipped.
           Parsing with Jinja/YAML is also skipped, if the argument value is a
           string that starts with '_noparse_', '_noparsejinja_',
           '_noparseyaml_', respectively.
           If argument 'name' does not exist, the function raises an
           AttributeError, unless a 'default' value is given.
           """

        def parse(arg_):

            # Recursively parse list items
            if isinstance(arg_, list):
                return [parse(item) for item in arg_]

            # Recursively parse dict values (not keys!)
            if isinstance(arg_, dict):
                return dict((key, parse(val)) for key, val in arg_.items())

            if isinstance(arg_, str):
                if parse_jinja and \
                   not isinstance(arg_, scriptengine.yaml.NoParseJinjaString):
                    # Make sure that a NoParseString is still a NoParseString
                    # after this!
                    arg_ = type(arg_)(scriptengine.jinja.render(arg_, context))

                if parse_yaml and \
                   not isinstance(arg_, scriptengine.yaml.NoParseYamlString):
                    try:
                        return yaml.full_load(arg_)
                    except (yaml.scanner.ScannerError,
                            yaml.parser.ParserError,
                            yaml.constructor.ConstructorError):
                        self.log_debug(f'Reparsing argument "{arg_}" '
                                       'with YAML failed')

                # Return plain strings, not NoParse*Strings
                return str(arg_)

            # If not a list, dict or string, just return
            return arg_

        try:
            arg = getattr(self, name)

        except AttributeError:
            if default is _SENTINEL:
                self.log_error(f'Missing task argument: {name}')
                raise ScriptEngineTaskArgumentMissingError(
                                        f'Missing task argument: {name}')
            arg = default

        return parse(arg)

    def _log(self, level, msg):
        logger = logging.getLogger('se.task')
        logger.log(level, msg, extra={'type': self.reg_name,
                                      'id': self.shortid})

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



"""Context task for ScriptEngine."""

import yaml
from deepmerge import always_merger

import scriptengine.jinja
from scriptengine.tasks.core import Task, timed_runner


class Context(Task):
    """Context task, sets parameters in the SE context

       The tasks run() method processes, recursively, a dictionary of key:value
       pairs and adds them to the context. The parameter values are rendered
       with Jinja2 and YAML unless they are of the special NoParse*String
       string types (see the YAML constructors defined in yaml.py).
    """

    @timed_runner
    def run(self, context):

        def evaluate_params(parameters, context):
            """Recursively evaluates parameter dict by parsing with Jinja2 and
               reparsing with YAML
            """
            evaluated_params = {}
            for key, val in parameters.items():

                if isinstance(val, str):
                    if not isinstance(val,
                                      scriptengine.yaml.NoParseJinjaString):
                        # Make sure that a NoParseString is still a
                        # NoParseString after this!
                        val = type(val)(scriptengine.jinja.render(val,
                                                                  context))

                    if not isinstance(val,
                                      scriptengine.yaml.NoParseYamlString):
                        # Try to reload the item through YAML, in order to
                        # get the right type
                        # Otherwise, everything would be just a string
                        try:
                            val = yaml.full_load(val)
                        # However, it may really be a string that is
                        # no valid YAML!
                        except (yaml.parser.ParserError,
                                yaml.constructor.ConstructorError):
                            self.log_debug(f'Reparsing argument "{val}" '
                                           'with YAML failed')

                    # For consistency, we always return plain strings
                    if isinstance(val, scriptengine.yaml.NoParseString):
                        val = str(val)

                    evaluated_params[key] = val

                elif isinstance(val, dict):
                    evaluated_params[key] = evaluate_params(val, context)

                else:
                    evaluated_params[key] = val

            return evaluated_params

        params = {key: val for key, val in self.__dict__.items()
                  if not (isinstance(key, str) and key.startswith('_'))}
        self.log_info(f'{", ".join([f"{k}={params[k]}" for k in params])}')

        add_to_context = evaluate_params(params, context)
        self.log_debug(f'Adding to context: {add_to_context}')
        always_merger.merge(context, add_to_context)



"""Command task for ScriptEngine

   Example:
      - command:
            name: ls
            args: [-l, -a]
"""

import subprocess

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.exceptions import ScriptEngineTaskRunError


class Command(Task):
    """Command task, executes a command in a shell"""

    _required_arguments = ('name', )

    def __init__(self, arguments):
        Command.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        self.log_info(f"{self.name} "
                      f"args={getattr(self, 'args', 'none')} "
                      f"cwd={getattr(self, 'cwd', 'none')}")

        command = self.getarg('name', context)

        args = self.getarg('args', context, default=[])
        args = args if isinstance(args, list) else [args]

        cwd = self.getarg('cwd', context, default=None)

        self.log_debug(f"{command} {' '.join(map(str,args))}"
                       f"{' cwd='+cwd if cwd else ''}")

        stdout = self.getarg('stdout', default=None)
        try:
            result = subprocess.run(
                        map(str, [command]+args),
                        stdout=subprocess.PIPE if stdout is not None else None,
                        cwd=cwd,
                        check=True)
        except subprocess.CalledProcessError as error:
            msg = (f'Command "{self.name}" returned '
                   f'error code {error.returncode}')
            if self.getarg('ignore_error', context, default=False):
                self.log_warning(msg)
            else:
                self.log_error(msg)
                raise ScriptEngineTaskRunError(msg)
        else:
            if stdout is not None:
                context[stdout] = [binary.decode("UTF-8")
                                   for binary in result.stdout.splitlines()]



"""ScripteEngine helpers: Colored terminal output

Terminal markup and colors by using ANSI escape sequences. To print colored
text, use, for example:

    import terminal_colors as TC

    print(f'{TC.RED}Hello!{TC.RESET}')

For background colors, use BG_ prefix, i.e. TC.BG_GREEN.

The module supports color themes. To set another theme, use, for example

    TC.set_theme('none')

Further output will then be without colors.
"""

import sys

from scriptengine.exceptions import ScriptEngineError

# The default color theme
ACTIVE_THEME = 'standard'

# All themes with all colors.
# It is not necessary that all themes define the same colors. But if you are
# using a color that is not defined in the active theme, there will be an
# AttributeError.
COLORTABLE = {
    'standard': {
        'RESET': '\033[0m',

        'BOLD': '\033[01m', 'DIM': '\033[02m', 'UNDERLINE': '\033[04m',
        'REVERSE': '\033[07m', 'STRIKETHROUGH': '\033[09m',
        'INVISIBLE': '\033[08m',

        'BLACK': '\033[30m', 'RED': '\033[31m', 'GREEN': '\033[32m',
        'ORANGE': '\033[33m', 'BLUE': '\033[34m', 'PURPLE': '\033[35m',
        'CYAN': '\033[36m', 'LIGHTGREY': '\033[37m', 'DARKGREY': '\033[90m',
        'LIGHTRED': '\033[91m', 'LIGHTGREEN': '\033[92m',
        'YELLOW': '\033[93m', 'LIGHTBLUE': '\033[94m', 'PINK': '\033[95m',
        'LIGHTCYAN': '\033[96m',

        'BG_BLACK': '\033[40m', 'BG_RED': '\033[41m', 'BG_GREEN': '\033[42m',
        'BG_ORANGE': '\033[43m', 'BG_BLUE': '\033[44m',
        'BG_PURPLE': '\033[45m', 'BG_CYAN': '\033[46m',
        'BG_LIGHTGREY': '\033[47m'},

    'none': {
        'RESET': '',

        'BOLD': '', 'DIM': '', 'UNDERLINE': '', 'REVERSE': '',
        'STRIKETHROUGH': '', 'INVISIBLE': '',

        'BLACK': '', 'RED': '', 'GREEN': '', 'ORANGE': '',
        'BLUE': '', 'PURPLE': '', 'CYAN': '', 'LIGHTGREY': '',
        'DARKGREY': '', 'LIGHTRED': '', 'LIGHTGREEN': '',
        'YELLOW': '', 'LIGHTBLUE': '', 'PINK': '',
        'LIGHTCYAN': '',

        'BG_BLACK': '', 'BG_RED': '', 'BG_GREEN': '',
        'BG_ORANGE': '', 'BG_BLUE': '', 'BG_PURPLE': '',
        'BG_CYAN': '', 'BG_LIGHTGREY': ''}
    }


def set_theme(theme):
    """ Changes the active color theme
    """
    global ACTIVE_THEME
    if theme not in COLORTABLE.keys():
        raise ScriptEngineError(f'Invalid terminal color scheme: "{theme}"')
    ACTIVE_THEME = theme

    # Clean all present color settings
    for color in {c for t in COLORTABLE for c in COLORTABLE[t]}:
        if hasattr(sys.modules[__name__], color):
            delattr(sys.modules[__name__], color)

    # Set colors for new theme
    for color, value in COLORTABLE[theme].items():
        setattr(sys.modules[__name__], color, value)


# Set default colors
for _color, _value in COLORTABLE[ACTIVE_THEME].items():
    setattr(sys.modules[__name__], _color, _value)



"""Template task for ScriptEngine."""

import os
import jinja2

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.jinja import render as j2render
from scriptengine.jinja import filters as j2filters


class Template(Task):
    """Template task, renders a template with Jinja2 and writes result to file.

    The task needs a source (a Jinja2 template file) and a destination. It
    renders the template given the context passed in kwargs and writes the
    result to the destination file.

    Args:
        dictionary (dict): Must at least contain the following keys:
            - src: Source file name (a Jinja2 template)
            - dst: Destination file name
    """
    _required_arguments = ('src', 'dst', )

    def __init__(self, arguments):
        Template.check_arguments(arguments)
        super().__init__(arguments)

    def __str__(self):
        return f'Template: {self.src} --> {self.dst}'

    @timed_runner
    def run(self, context):
        src = self.getarg('src', context)
        dst = self.getarg('dst', context)
        self.log_info(f'Render {src} --> {dst}')

        # The template search path:
        #   1. .
        #   2. ./templates
        #   3. <cwd>
        #   4. <cwd>/templates
        # where <cwd> is the original working directory at the time when the
        # se script was called
        search_path = ('.',
                       './templates',
                       context.se.cli.cwd,
                       os.path.join(context.se.cli.cwd, 'templates'))
        self.log_debug(f'Search path for template: {search_path}')

        loader = jinja2.FileSystemLoader(search_path)
        environment = jinja2.Environment(loader=loader)
        for name, function in j2filters().items():
            environment.filters[name] = function
        template = environment.get_template(src)
        output_text = j2render(template.render(context), context)

        with open(dst, 'w') as output_file:
            output_file.write(f'{output_text}\n')
